from xarray import DataArray

import weldx.asdf.tags.weldx.core.common_types as ct
from weldx.asdf.types import WeldxType


class XarrayDataArrayASDF(WeldxType):
    """Serialization class for xarray.DataArray"""

    name = "core/data_array"
    version = "1.0.0"
    types = [DataArray]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: DataArray, ctx):
        """
        Convert an instance of the 'xarray.DataArray' type into YAML  representations.

        Parameters
        ----------
        node :
            Instance of the 'xarray.DataArray' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'xarray.DataArray'
            type to be serialized.

        """
        attributes = node.attrs
        coordinates = []
        data = ct.Variable("data", node.dims, node.data)

        for name, coord_data in node.coords.items():
            coordinates.append(ct.Variable(name, coord_data.dims, coord_data.data))

        tree = {
            "attributes": attributes,
            "coordinates": coordinates,
            "data": data,
        }

        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into an 'xarray.DataArray'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        xarray.DataArray :
            An instance of the 'xarray.DataArray' type.

        """
        data = tree["data"].data
        dims = tree["data"].dimensions
        coords = {}
        for coordinate in tree["coordinates"]:
            coords[coordinate.name] = (coordinate.dimensions, coordinate.data)

        obj = DataArray(data=data, coords=coords, dims=dims)

        obj.attrs = tree["attributes"]

        return obj



from dataclasses import dataclass

from weldx.asdf.types import WeldxType
from weldx.asdf.utils import drop_none_attr

__all__ = ["Workpiece", "WorkpieceType"]


@dataclass
class Workpiece:
    """<CLASS DOCSTRING>"""

    geometry: str


class WorkpieceType(WeldxType):
    """<ASDF TYPE DOCSTRING>"""

    name = "aws/design/workpiece"
    version = "1.0.0"
    types = [Workpiece]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: Workpiece, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = drop_none_attr(node)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = Workpiece(**tree)
        return obj



"""Test the measurement package."""

import sympy
import xarray as xr

import weldx.measurement as msm
from weldx.asdf.utils import _write_read_buffer
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.core import MathematicalExpression


def test_generic_measurement():
    """Test basic measurement creation and ASDF read/write."""
    data_01 = msm.Data(
        name="Welding current", data=xr.DataArray([1, 2, 3, 4], dims=["time"])
    )

    data_02 = msm.Data(
        name="Welding voltage", data=xr.DataArray([10, 20, 30, 40], dims=["time"])
    )

    src_01 = msm.Source(
        name="Current Sensor",
        output_signal=msm.Signal("analog", "V", data=None),
        error=msm.Error(1337.42),
    )

    src_02 = msm.Source(
        name="Voltage Sensor",
        output_signal=msm.Signal("analog", "V", data=None),
        error=msm.Error(1),
    )

    dp_01 = msm.DataTransformation(
        name="AD conversion current measurement",
        input_signal=src_01.output_signal,
        output_signal=msm.Signal("digital", "V", data=None),
        error=msm.Error(999.0),
    )

    dp_02 = msm.DataTransformation(
        name="Calibration current measurement",
        input_signal=dp_01.output_signal,
        output_signal=msm.Signal("digital", "A", data=data_01),
        error=msm.Error(43.0),
    )

    dp_03 = msm.DataTransformation(
        name="AD conversion voltage measurement",
        input_signal=dp_02.output_signal,
        output_signal=msm.Signal("digital", "V", data=None),
        error=msm.Error(2.0),
    )

    dp_04 = msm.DataTransformation(
        name="Calibration voltage measurement",
        input_signal=dp_03.output_signal,
        output_signal=msm.Signal("digital", "V", data=data_02),
        error=msm.Error(Q_(3.0, "percent")),
    )

    chn_01 = msm.MeasurementChain(
        name="Current measurement", data_source=src_01, data_processors=[dp_01, dp_02]
    )

    chn_02 = msm.MeasurementChain(
        name="Voltage measurement", data_source=src_02, data_processors=[dp_03, dp_04]
    )

    eqp_01 = msm.GenericEquipment(
        "Current Sensor", sources=[src_01], data_transformations=[dp_02]
    )
    eqp_02 = msm.GenericEquipment(
        "AD Converter", sources=None, data_transformations=[dp_01, dp_03]
    )
    eqp_03 = msm.GenericEquipment(
        "Voltage Sensor", sources=None, data_transformations=[dp_04]
    )

    measurement_01 = msm.Measurement(
        name="Current measurement", data=[data_01], measurement_chain=chn_01
    )
    measurement_02 = msm.Measurement(
        name="Voltage measurement", data=[data_02], measurement_chain=chn_02
    )

    equipment = [eqp_01, eqp_02, eqp_03]
    measurement_data = [data_01, data_02]
    measurement_chains = [chn_01]
    measurements = [measurement_01, measurement_02]
    sources = [src_01]
    processors = [dp_01, dp_02]

    [a, x, b] = sympy.symbols("a x b")
    expr_01 = MathematicalExpression(a * x + b)
    expr_01.set_parameter("a", 2)
    expr_01.set_parameter("b", 3)
    print(expr_01.parameters)
    print(expr_01.get_variable_names())
    print(expr_01.evaluate(x=3))

    tree = {
        "equipment": equipment,
        "data": measurement_data,
        "measurements": measurements,
        # "expression": expr_01,
        "measurement_chains": measurement_chains,
        "data_sources": sources,
        "data_processors": processors,
    }

    _write_read_buffer(tree)



"""Test the internal utility functions."""

import math

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from pandas import DatetimeIndex as DTI
from pandas import TimedeltaIndex as TDI
from pandas import date_range
from pint.errors import DimensionalityError

import weldx.utility as ut
from weldx.constants import WELDX_QUANTITY as Q_


def test_is_column_in_matrix():
    """Test the is_column_in_matrix function.

    Test should be self explanatory.

    """
    c_0 = [1, 5, 2]
    c_1 = [3, 2, 2]
    c_2 = [1, 6, 1]
    c_3 = [1, 6, 0]
    matrix = np.array([c_0, c_1, c_2, c_3]).transpose()

    assert ut.is_column_in_matrix(c_0, matrix)
    assert ut.is_column_in_matrix(c_1, matrix)
    assert ut.is_column_in_matrix(c_2, matrix)
    assert ut.is_column_in_matrix(c_3, matrix)

    assert not ut.is_column_in_matrix([1, 6], matrix)
    assert not ut.is_column_in_matrix([1, 6, 2], matrix)
    assert not ut.is_column_in_matrix([1, 1, 3, 1], matrix)


def test_is_row_in_matrix():
    """Test the is_row_in_matrix function.

    Test should be self explanatory.

    """
    c_0 = [1, 5, 2]
    c_1 = [3, 2, 2]
    c_2 = [1, 6, 1]
    c_3 = [1, 6, 0]
    matrix = np.array([c_0, c_1, c_2, c_3])

    assert ut.is_row_in_matrix(c_0, matrix)
    assert ut.is_row_in_matrix(c_1, matrix)
    assert ut.is_row_in_matrix(c_2, matrix)
    assert ut.is_row_in_matrix(c_3, matrix)

    assert not ut.is_row_in_matrix([1, 6], matrix)
    assert not ut.is_row_in_matrix([1, 6, 2], matrix)
    assert not ut.is_row_in_matrix([1, 1, 3, 1], matrix)


def test_matrix_is_close():
    """Test the matrix_is_close function.

    Test should be self explanatory.

    """
    mat_a = np.array([[0, 1, 2], [3, 4, 5]])
    mat_b = np.array([[3, 5, 1], [7, 1, 9]])

    assert ut.matrix_is_close(mat_a, mat_a)
    assert ut.matrix_is_close(mat_b, mat_b)
    assert not ut.matrix_is_close(mat_a, mat_b)
    assert not ut.matrix_is_close(mat_b, mat_a)

    # check tolerance
    mat_c = mat_a + 0.0001
    assert ut.matrix_is_close(mat_a, mat_c, abs_tol=0.00011)
    assert not ut.matrix_is_close(mat_a, mat_c, abs_tol=0.00009)

    # vectors have different size
    assert not ut.matrix_is_close(mat_a, mat_a[0:2, 0:2])


def test_vector_is_close():
    """Test the vector_is_close function.

    Test should be self explanatory.

    """
    vec_a = np.array([0, 1, 2])
    vec_b = np.array([3, 5, 1])

    assert ut.vector_is_close(vec_a, vec_a)
    assert ut.vector_is_close(vec_b, vec_b)
    assert not ut.vector_is_close(vec_a, vec_b)
    assert not ut.vector_is_close(vec_b, vec_a)

    # check tolerance
    vec_c = vec_a + 0.0001
    assert ut.vector_is_close(vec_a, vec_c, abs_tol=0.00011)
    assert not ut.vector_is_close(vec_a, vec_c, abs_tol=0.00009)

    # vectors have different size
    assert not ut.vector_is_close(vec_a, vec_a[0:2])


@pytest.mark.parametrize(
    "arg, expected",
    [
        # timedeltas
        (TDI([42]), TDI([42])),
        (pd.timedelta_range("0s", "20s", 10), pd.timedelta_range("0s", "20s", 10)),
        (np.timedelta64(42), TDI([42])),
        (np.array([-10, 0, 20]).astype(np.timedelta64), TDI([-10, 0, 20])),
        (Q_(42, "ns"), TDI([42])),
        ("10s", TDI(["10s"])),
        (["5ms", "10s", "2D"], TDI(["5 ms", "10s", "2D"])),
        # datetimes
        (np.datetime64(50, "Y"), DTI(["2020-01-01"])),
        ("2020-01-01", DTI(["2020-01-01"])),
        (
            np.array(
                ["2012-10-02", "2012-10-05", "2012-10-11"], dtype="datetime64[ns]"
            ),
            DTI(["2012-10-02", "2012-10-05", "2012-10-11"]),
        ),
    ],
)
def test_to_pandas_time_index(arg, expected):
    """Test conversion to appropriate pd.TimedeltaIndex or pd.DatetimeIndex."""
    assert np.all(ut.to_pandas_time_index(arg) == expected)


@pytest.mark.parametrize(
    "arg, exception",
    [(5, TypeError), ("string", TypeError), (Q_(10, "m"), DimensionalityError)],
)
def test_to_pandas_time_index_exceptions(arg, exception):
    """Test correct exceptions on invalid inputs."""
    with pytest.raises(exception):
        ut.to_pandas_time_index(arg)


def test_pandas_time_delta_to_quantity():
    """Test the 'pandas_time_delta_to_quantity' utility function."""
    is_close = np.vectorize(math.isclose)

    def _check_close(t1, t2):
        assert np.all(is_close(t1.magnitude, t2.magnitude))
        assert t1.units == t2.units

    time_single = pd.TimedeltaIndex([1], unit="s")

    _check_close(ut.pandas_time_delta_to_quantity(time_single), Q_(1, "s"))
    _check_close(ut.pandas_time_delta_to_quantity(time_single, "ms"), Q_(1000, "ms"))
    _check_close(ut.pandas_time_delta_to_quantity(time_single, "us"), Q_(1000000, "us"))
    _check_close(
        ut.pandas_time_delta_to_quantity(time_single, "ns"), Q_(1000000000, "ns")
    )

    time_multi = pd.TimedeltaIndex([1, 2, 3], unit="s")
    _check_close(ut.pandas_time_delta_to_quantity(time_multi), Q_([1, 2, 3], "s"))
    _check_close(
        ut.pandas_time_delta_to_quantity(time_multi, "ms"), Q_([1000, 2000, 3000], "ms")
    )
    _check_close(
        ut.pandas_time_delta_to_quantity(time_multi, "us"),
        Q_([1000000, 2000000, 3000000], "us"),
    )
    _check_close(
        ut.pandas_time_delta_to_quantity(time_multi, "ns"),
        Q_([1000000000, 2000000000, 3000000000], "ns"),
    )


def test_xr_interp_like():
    """Test behaviour of custom interpolation method for xarray Objects."""
    # basic interpolation behavior on a single coordinate
    n_a = 5  # range of "a" coordinate in da_a
    s_a = 0.5  # default steps in "a" coordinate in da_a
    da_a = xr.DataArray(
        np.arange(0, n_a + s_a, s_a),
        dims=["a"],
        coords={"a": np.arange(0, n_a + s_a, s_a)},
    )

    # interp from subset inside original data
    test = ut.xr_interp_like(da_a.loc[2:4:2], da_a)
    assert test.a[0] == da_a.a[0]
    assert test.a[-1] == da_a.a[-1]
    assert test[0] == da_a.loc[2]
    assert test[-1] == da_a.loc[4]

    # interp with overlap
    test = ut.xr_interp_like(da_a.loc[1:3], da_a.loc[2:4])
    assert test.a[0] == da_a.a.loc[2]
    assert test.a[-1] == da_a.a.loc[4]
    assert test[0] == da_a.loc[2]
    assert test[-1] == da_a.loc[3]
    assert np.all(test.loc[3:4] == da_a.loc[3])

    # overlap without fill (expecting nan values for out of range indexes)
    test = ut.xr_interp_like(da_a.loc[1:3], da_a.loc[2:4], fillna=False)
    assert test.a[0] == da_a.a.loc[2]
    assert test.a[-1] == da_a.a.loc[4]
    assert test[0] == da_a.loc[2]
    assert np.isnan(test[-1])
    assert np.all(np.isnan(test.where(test.a > 3, drop=True)))

    # outside interpolation without overlap
    test = ut.xr_interp_like(da_a.loc[0:3], da_a.loc[4:5])
    assert test.a[0] == da_a.a.loc[4]
    assert test.a[-1] == da_a.a.loc[5]
    assert np.all(test.loc[4:5] == da_a.loc[3])

    # single point to array interpolation
    # important: da_a.loc[5] for indexing would drop coordinates (unsure why)
    test = ut.xr_interp_like(da_a.loc[2:2], da_a)
    assert np.all(test.a == da_a.a)
    assert np.all(test == da_a.loc[2:2])
    with pytest.raises(ValueError):
        ut.xr_interp_like(da_a.loc[2:2], da_a, fillna=False)

    # single point to single point interpolation (different points)
    test = ut.xr_interp_like(da_a.loc[2:2], da_a.loc[3:3])
    assert np.all(test.a == da_a.a)
    assert np.all(test == da_a.loc[2:2])

    # single point to single point interpolation (matching points)
    test = ut.xr_interp_like(da_a.loc[2:2], da_a.loc[2:2])
    assert np.all(test.a == da_a.a)
    assert np.all(test == da_a.loc[2:2])

    # dict-like inputs on existing coordinate
    test = ut.xr_interp_like(da_a, {"a": np.arange(-5, 15, 0.5)})
    assert np.all(test.where(test.a < da_a.a.min(), drop=True) == da_a.min())
    assert np.all(test.where(test.a > da_a.a.max(), drop=True) == da_a[-1])
    assert np.all(test.where(test.a.isin(da_a.a), drop=True) == da_a)

    da = da_a.loc[3:3]
    test = ut.xr_interp_like(da, {"a": np.arange(-5, 15, 0.5)})
    assert np.all(test.where(test.a < da.a.min(), drop=True) == da.min())
    assert np.all(test.where(test.a > da.a.max(), drop=True) == da[-1])
    assert np.all(test.where(test.a.isin(da.a), drop=True) == da)

    test = ut.xr_interp_like(da_a, {"a": np.arange(-5, 15, 0.5)}, fillna=False)
    assert np.all(np.isnan((test.where(test.a < 0, drop=True))))
    assert np.all(np.isnan(test.where(test.a > da_a.a.max(), drop=True)))
    assert np.all(test.where(test.a.isin(da_a.a), drop=True) == da_a)

    # dict-like inputs on new coordinate with/without broadcasting
    da1 = da_a
    test = ut.xr_interp_like(da1, {"b": np.arange(3)})
    assert test.equals(da_a)

    da1 = da_a
    test = ut.xr_interp_like(da1, {"b": np.arange(3)}, broadcast_missing=True)
    assert test.equals(da1.broadcast_like(test))

    # test coordinate selection with interp_coords
    da1 = da_a
    test = ut.xr_interp_like(
        da1,
        {"b": np.arange(3), "c": np.arange(3)},
        broadcast_missing=True,
        interp_coords=["c"],
    )
    assert "b" not in test.coords
    assert "c" in test.coords

    # catch error on unsorted array
    da = xr.DataArray([0, 1, 2, 3], dims="a", coords={"a": [2, 1, 3, 0]})
    with pytest.raises(ValueError):
        test = ut.xr_interp_like(da, {"a": np.arange(6)}, assume_sorted=True)

    # basic interpolation behavior with different coordinates (broadcasting)
    n_b = 3  # range of "b" coordinate in da_b
    s_b = 1  # default steps in "b" coordinate in da_b
    da_b = xr.DataArray(
        np.arange(0, n_b + s_b, s_b) ** 2,
        dims=["b"],
        coords={"b": np.arange(0, n_b + s_b, s_b)},
    )

    assert da_a.equals(ut.xr_interp_like(da_a, da_b))
    assert da_a.broadcast_like(da_b).broadcast_equals(
        ut.xr_interp_like(da_a, da_b, broadcast_missing=True)
    )

    # coords syntax
    assert da_a.broadcast_like(da_b).broadcast_equals(
        ut.xr_interp_like(da_a, da_b.coords, broadcast_missing=True)
    )

    # sorting and interpolation with multiple dimensions
    a = np.arange(3, 6)
    b = np.arange(1, -3, -1)
    da_ab = xr.DataArray(
        a[..., np.newaxis] @ b[np.newaxis, ...],
        dims=["a", "b"],
        coords={"a": a, "b": b},
    )

    a_new = np.arange(3, 5, 0.5)
    b_new = np.arange(-1, 1, 0.5)
    test = ut.xr_interp_like(
        da_ab,
        {"a": a_new, "b": b_new, "c": np.arange(2)},
        assume_sorted=False,
        broadcast_missing=True,
    )
    assert np.all(
        test.transpose(..., "a", "b") == a_new[..., np.newaxis] @ b_new[np.newaxis, ...]
    )

    # tests with time data types
    # TODO: add more complex test examples
    t = pd.timedelta_range(start="10s", end="0s", freq="-1s", closed="left")
    da_t = xr.DataArray(np.arange(10, 0, -1), dims=["t"], coords={"t": t})

    test = ut.xr_interp_like(
        da_t,
        {"t": pd.timedelta_range(start="3s", end="7s", freq="125ms", closed="left")},
    )
    assert np.all(test == np.arange(3, 7, 0.125))


@pytest.mark.parametrize(
    "list_of_objects, time_exp",
    [
        (
            [
                date_range("2020-02-02", periods=4, freq="2D"),
                date_range("2020-02-01", periods=4, freq="2D"),
                date_range("2020-02-03", periods=2, freq="3D"),
            ],
            date_range("2020-02-01", periods=8, freq="1D"),
        ),
        ([TDI([1, 5]), TDI([2, 6, 7]), TDI([1, 3, 7])], TDI([1, 2, 3, 5, 6, 7])),
    ],
)
def test_get_time_union(list_of_objects, time_exp):
    """Test input types for get_time_union function.

    Parameters
    ----------
    list_of_objects:
        List with input objects
    time_exp:
        Expected result time

    """
    assert np.all(ut.get_time_union(list_of_objects) == time_exp)


def test_xf_fill_all():
    """Test filling along all dimensions."""
    da1 = xr.DataArray(
        np.eye(2), dims=["a", "b"], coords={"a": np.arange(2), "b": np.arange(2)}
    )
    da2 = xr.DataArray(
        dims=["a", "b"], coords={"a": np.arange(-1, 3, 0.5), "b": np.arange(-1, 3, 0.5)}
    )
    da3 = da1.broadcast_like(da2)

    da4 = ut.xr_fill_all(da3, order="fb")
    assert not np.any(np.isnan(da4))
    assert np.all(da4[0:4, 0:4] == np.ones((4, 4)))
    assert np.all(da4[0:4, 4:8] == np.zeros((4, 4)))
    assert np.all(da4[4:8, 0:4] == np.zeros((4, 4)))
    assert np.all(da4[4:8, 4:8] == np.ones((4, 4)))

    da4 = ut.xr_fill_all(da3, order="bf")
    assert not np.any(np.isnan(da4))
    assert np.all(da4[0:3, 0:3] == np.ones((3, 3)))
    assert np.all(da4[0:3, 3:8] == np.zeros((3, 5)))
    assert np.all(da4[3:8, 0:3] == np.zeros((5, 3)))
    assert np.all(da4[3:8, 3:8] == np.ones((5, 5)))

    with pytest.raises(ValueError):
        ut.xr_fill_all(da3, order="wrong")


_dax_check = xr.DataArray(
    data=np.ones((2, 2, 2, 4, 3)),
    dims=["d1", "d2", "d3", "d4", "d5"],
    coords={
        "d1": np.array([-1, 1], dtype=float),
        "d2": np.array([-1, 1], dtype=int),
        "d3": pd.DatetimeIndex(["2020-05-01", "2020-05-03"]),
        "d4": pd.TimedeltaIndex([0, 1, 2, 3], "s"),
        "d5": ["x", "y", "z"],
    },
)

_dax_ref = dict(
    d1={"values": np.array([-1, 1]), "dtype": "float"},
    d2={"values": np.array([-1, 1]), "dtype": int},
    d3={
        "values": pd.DatetimeIndex(["2020-05-01", "2020-05-03"]),
        "dtype": ["datetime64[ns]", "timedelta64[ns]"],
    },
    d4={
        "values": pd.TimedeltaIndex([0, 1, 2, 3], "s"),
        "dtype": ["datetime64[ns]", "timedelta64[ns]"],
    },
    d5={"values": ["x", "y", "z"], "dtype": "<U1"},
)


@pytest.mark.parametrize(
    "dax, ref_dict",
    [
        (_dax_check, _dax_ref),
        (_dax_check.coords, _dax_ref),
        (_dax_check, {"d1": {"dtype": ["float64", int]}}),
        (_dax_check, {"d2": {"dtype": ["float64", int]}}),
        (_dax_check, {"no_dim": {"optional": True, "dtype": float}}),
        (_dax_check, {"d5": {"dtype": str}}),
        (_dax_check, {"d5": {"dtype": [str]}}),
        (_dax_check, {"d4": {"dtype": "timedelta64"}}),
        (_dax_check, {"d3": {"dtype": ["datetime64", "timedelta64"]}}),
    ],
)
def test_xr_check_coords(dax, ref_dict):
    """Test weldx.utility.xr_check_coords function."""
    assert ut.xr_check_coords(dax, ref_dict)


@pytest.mark.parametrize(
    "dax, ref_dict, exception_type",
    [
        (_dax_check, {"d1": {"dtype": int}}, TypeError),
        (_dax_check, {"d1": {"dtype": int, "optional": True}}, TypeError),
        (_dax_check, {"no_dim": {"dtype": float}}, KeyError),
        (
            _dax_check,
            {"d5": {"values": ["x", "noty", "z"], "dtype": "str"}},
            ValueError,
        ),
        (_dax_check, {"d1": {"dtype": [int, str, bool]}}, TypeError),
        (_dax_check, {"d4": {"dtype": "datetime64"}}, TypeError),
        ({"d4": np.arange(4)}, {"d4": {"dtype": "int"}}, ValueError),
    ],
)
def test_xr_check_coords_exception(dax, ref_dict, exception_type):
    """Test weldx.utility.xr_check_coords function."""
    with pytest.raises(exception_type):
        ut.xr_check_coords(dax, ref_dict)



"""Debug asdf save of groove implementation (run after groove_save_script.py)."""

import asdf

from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension
from weldx.asdf.tags.weldx.core.iso_groove import (
    DHUGroove,
    DHVGroove,
    DUGroove,
    DVGroove,
    FFGroove,
    HUGroove,
    HVGroove,
    IGroove,
    UGroove,
    UVGroove,
    VGroove,
    VVGroove,
)

opened = asdf.open("testfile.yml", extensions=[WeldxAsdfExtension(), WeldxExtension()])

if isinstance(opened["test001"], VGroove):
    print(opened["test001"])
    opened["test001"].plot()
else:
    print("false")

if isinstance(opened["test002"], UGroove):
    print(opened["test002"])
    opened["test002"].plot()
else:
    print("false")

if isinstance(opened["test003"], IGroove):
    print(opened["test003"])
    opened["test003"].plot()
else:
    print("false")

if isinstance(opened["test004"], UVGroove):
    print(opened["test004"])
    opened["test004"].plot()
else:
    print("false")

if isinstance(opened["test005"], VVGroove):
    print(opened["test005"])
    opened["test005"].plot()
else:
    print("false")

if isinstance(opened["test006"], HVGroove):
    print(opened["test006"])
    opened["test006"].plot()
else:
    print("false")

if isinstance(opened["test007"], HUGroove):
    print(opened["test007"])
    opened["test007"].plot()
else:
    print("false")

if isinstance(opened["test008"], DVGroove):
    print(opened["test008"])
    opened["test008"].plot()
else:
    print("false")

if isinstance(opened["test009"], DUGroove):
    print(opened["test009"])
    opened["test009"].plot()
else:
    print("false")

if isinstance(opened["test010"], DHVGroove):
    print(opened["test010"])
    opened["test010"].plot()
else:
    print("false")

if isinstance(opened["test011"], DHUGroove):
    print(opened["test011"])
    opened["test011"].plot()
else:
    print("false")

if isinstance(opened["test012"], FFGroove):
    print(opened["test012"])
    opened["test012"].plot()
else:
    print("false")

if isinstance(opened["test013"], FFGroove):
    print(opened["test013"])
    opened["test013"].plot()
else:
    print("false")

if isinstance(opened["test014"], FFGroove):
    print(opened["test014"])
    opened["test014"].plot()
else:
    print("false")

if isinstance(opened["test015"], FFGroove):
    print(opened["test015"])
    opened["test015"].plot()
else:
    print("false")

if isinstance(opened["test016"], FFGroove):
    print(opened["test016"])
    opened["test016"].plot()
else:
    print("false")

if isinstance(opened["test017"], FFGroove):
    print(opened["test017"])
    opened["test017"].plot()
else:
    print("false")



# Licensed under a 3-clause BSD style license - see LICENSE
# -*- coding: utf-8 -*-
"""Python Library description and install setup."""

from os import path

from setuptools import find_packages, setup

import versioneer

# add README.md as long description
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy>=0.18",
    "pandas>=1.0",
    "xarray>=0.15",
    "scipy>=1.4",
    "sympy>=1.6",
    "pint>=0.11",
    "asdf>=2.7",
    "bottleneck>=1.3",
    "Jinja2>=2.11",
    "networkx>=2",
    "matplotlib>=3",
]

entry_points = {}
entry_points["asdf_extensions"] = [
    "weldx = weldx.asdf.extension:WeldxExtension",
    "weldx-asdf = weldx.asdf.extension:WeldxAsdfExtension",
]

setup(
    name="weldx",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Cagtay Fabry",
    author_email="Cagtay.Fabry@bam.de",
    packages=find_packages(),
    url="https://www.bam.de/weldx",
    license="BSD License",
    description="Python API for the WelDX file format and standard",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
        "Natural Language :: English",
    ],
    install_requires=requirements,
    include_package_data=True,  # include non-py files listed in MANIFEST.in
    entry_points=entry_points,  # register ASDF Extension entry_points
)



import sympy

from weldx.asdf.types import WeldxType
from weldx.core import MathematicalExpression

__all__ = ["MathematicalExpression", "MathematicalExpressionType"]


class MathematicalExpressionType(WeldxType):
    """Serialization class for sympy style math expressions."""

    name = "core/mathematical_expression"
    version = "1.0.0"
    types = [MathematicalExpression]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: MathematicalExpression, ctx):
        """convert to tagged tree and remove all None entries from node dictionary"""
        tree = {"expression": node.expression.__str__(), "parameters": node.parameters}
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        obj = MathematicalExpression(
            sympy.sympify(tree["expression"]), parameters=tree["parameters"]
        )
        return obj



"""Contains package internal utility functions."""

import math
from collections.abc import Iterable
from functools import reduce
from operator import or_
from typing import Any, Dict, List, Union

import numpy as np
import pandas as pd
import pint
import xarray as xr
from pandas.api.types import is_datetime64_dtype, is_object_dtype, is_timedelta64_dtype
from scipy.spatial.transform import Rotation as Rot
from scipy.spatial.transform import Slerp

import weldx.transformations as tf
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.core import MathematicalExpression, TimeSeries


def _sine(
    f: pint.Quantity,
    amp: pint.Quantity,
    bias: pint.Quantity = None,
    phase: pint.Quantity = Q_(0, "rad"),
) -> TimeSeries:  # pragma: no cover
    """Create a simple sine TimeSeries from quantity parameters.

    f(t) = amp*sin(f*t+phase)+bias

    Parameters
    ----------
    f : pint.Quantity
        Frequency of the sine (in Hz)
    amp : pint.Quantity
        Sine amplitude
    bias : pint.Quantity
        function bias
    phase : pint.Quantity
        phase shift

    Returns
    -------
    TimeSeries

    """
    if bias is None:
        bias = 0.0 * amp.u
    expr_string = "a*sin(o*t+p)+b"
    parameters = {"a": amp, "b": bias, "o": Q_(2 * np.pi, "rad") * f, "p": phase}
    expr = MathematicalExpression(expression=expr_string, parameters=parameters)
    return TimeSeries(expr)


def is_column_in_matrix(column, matrix) -> bool:
    """Check if a column (1d array) can be found inside of a matrix.

    Parameters
    ----------
    column :
        Column that should be checked
    matrix :
        Matrix

    Returns
    -------
    bool
        True or False

    """
    return is_row_in_matrix(column, np.transpose(matrix))


def is_row_in_matrix(row, matrix) -> bool:
    """Check if a row (1d array) can be found inside of a matrix.

    source: https://codereview.stackexchange.com/questions/193835

    Parameters
    ----------
    row :
        Row that should be checked
    matrix :
        Matrix

    Returns
    -------
    bool
        True or False

    """
    if not matrix.shape[1] == np.array(row).size:
        return False
    # noinspection PyUnresolvedReferences
    return (matrix == row).all(axis=1).any()


def to_float_array(container) -> np.ndarray:
    """Cast the passed container to a numpy array of floats.

    Parameters
    ----------
    container :
        Container which can be cast to a numpy array

    Returns
    -------
    numpy.ndarray

    """
    return np.array(container, dtype=float)


def to_list(var) -> list:
    """Store the passed variable into a list and return it.

    If the variable is already a list, it is returned without modification.
    If `None` is passed, the function returns an empty list.

    Parameters
    ----------
    var :
        Arbitrary variable

    Returns
    -------
    list

    """
    if isinstance(var, list):
        return var
    if var is None:
        return []
    return [var]


def to_pandas_time_index(
    time: Union[
        pint.Quantity,
        np.ndarray,
        pd.TimedeltaIndex,
        pd.DatetimeIndex,
        xr.DataArray,
        "tf.LocalCoordinateSystem",
    ],
) -> Union[pd.TimedeltaIndex, pd.DatetimeIndex]:
    """Convert a time variable to the corresponding pandas time index type.

    Parameters
    ----------
    time :
        Variable that should be converted.

    Returns
    -------
    pandas.TimedeltaIndex or pandas.DatetimeIndex
        Time union of all input objects

    """
    _input_type = type(time)

    if isinstance(time, (pd.DatetimeIndex, pd.TimedeltaIndex)):
        return time

    if isinstance(time, tf.LocalCoordinateSystem):
        return to_pandas_time_index(time.time)

    if isinstance(time, pint.Quantity):
        base = "s"  # using low base unit could cause rounding errors
        if not np.iterable(time):  # catch zero-dim arrays
            time = np.expand_dims(time, 0)
        return pd.TimedeltaIndex(data=time.to(base).magnitude, unit=base)

    if isinstance(time, (xr.DataArray, xr.Dataset)):
        if "time" in time.coords:
            time = time.time
        time_index = pd.Index(time.values)
        if is_timedelta64_dtype(time_index):
            if time.weldx.time_ref:
                time_index = time_index + time.weldx.time_ref
        return time_index

    if not np.iterable(time) or isinstance(time, str):
        time = [time]
    time = pd.Index(time)

    if isinstance(time, (pd.DatetimeIndex, pd.TimedeltaIndex)):
        return time

    # try manual casting for object dtypes (i.e. strings), should avoid integers
    # warning: this allows something like ["1","2","3"] which will be ns !!
    if is_object_dtype(time):
        for func in (pd.DatetimeIndex, pd.TimedeltaIndex):
            try:
                return func(time)
            except (ValueError, TypeError):
                continue

    raise TypeError(
        f"Could not convert {_input_type} " f"to pd.DatetimeIndex or pd.TimedeltaIndex"
    )


def pandas_time_delta_to_quantity(
    time: pd.TimedeltaIndex, unit: str = "s"
) -> pint.Quantity:
    """Convert a `pandas.TimedeltaIndex` into a corresponding `pint.Quantity`.

    Parameters
    ----------
    time : pandas.TimedeltaIndex
        Instance of `pandas.TimedeltaIndex`
    unit :
        String that specifies the desired time unit.

    Returns
    -------
    pint.Quantity :
        Converted time quantity

    """
    # from pandas Timedelta documentation: "The .value attribute is always in ns."
    # https://pandas.pydata.org/pandas-docs/version/0.23.4/generated/pandas.Timedelta.html
    nanoseconds = time.values.astype(np.int64)
    if len(nanoseconds) == 1:
        nanoseconds = nanoseconds[0]
    return Q_(nanoseconds, "ns").to(unit)


def matrix_is_close(mat_a, mat_b, abs_tol=1e-9) -> bool:
    """Check if a matrix is close or equal to another matrix.

    Parameters
    ----------
    mat_a :
        First matrix
    mat_b :
        Second matrix
    abs_tol :
        Absolute tolerance (Default value = 1e-9)

    Returns
    -------
    bool
        True or False

    """
    mat_a = to_float_array(mat_a)
    mat_b = to_float_array(mat_b)

    if not mat_a.shape == mat_b.shape:
        return False
    for i in range(mat_a.shape[0]):
        for j in range(mat_a.shape[1]):
            if not math.isclose(mat_a[i, j], mat_b[i, j], abs_tol=abs_tol):
                return False
    return True


def vector_is_close(vec_a, vec_b, abs_tol=1e-9) -> bool:
    """Check if a vector is close or equal to another vector.

    Parameters
    ----------
    vec_a :
        First vector
    vec_b :
        Second vector
    abs_tol :
        Absolute tolerance (Default value = 1e-9)

    Returns
    -------
    bool
        True or False

    """
    vec_a = to_float_array(vec_a)
    vec_b = to_float_array(vec_b)

    if not vec_a.size == vec_b.size:
        return False
    for i in range(vec_a.size):
        if not math.isclose(vec_a[i], vec_b[i], abs_tol=abs_tol):
            return False

    return True


def mat_vec_mul(a, b) -> np.ndarray:
    """Matrix, Vector multiplication using matmul with newaxis for correct broadcasting.

    Parameters
    ----------
    a :
        Input Matrix [m, n]
    b :
        Input Vector to be multiplied [n, ]

    Returns
    -------
    numpy.ndarray
        Resulting vector [n, ]

    """
    return np.matmul(a, b[..., np.newaxis]).squeeze()


def swap_list_items(arr, i1, i2) -> list:
    """Swap position of two items in a list.

    Parameters
    ----------
    arr :
        list in which to swap elements
    i1 :
        element 1 in list
    i2 :
        element 2 in list

    Returns
    -------
    list
        copy of list with swapped elements

    """
    i = list(arr).copy()
    a, b = i.index(i1), i.index(i2)
    i[b], i[a] = i[a], i[b]
    return i


def get_time_union(
    list_of_objects: List[Union[pd.DatetimeIndex, pd.TimedeltaIndex]]
) -> Union[pd.DatetimeIndex, pd.TimedeltaIndex]:
    """Generate a merged union of `pandas.DatetimeIndex` from list of inputs.

    The functions tries to merge common inputs that are "time-like" or might have time
    coordinates such as xarray objects, `~weldx.transformations.LocalCoordinateSystem`
    and other time objects. See `to_pandas_time_index` for supported input object types.

    Parameters
    ----------
    list_of_objects :
        list of input objects to merge

    Returns
    -------
    Union[pandas.DatetimeIndex, pandas.TimedeltaIndex]
        Pandas time index class with merged times

    """
    # TODO: add tests

    # see https://stackoverflow.com/a/44762908/11242411
    return reduce(or_, (to_pandas_time_index(idx) for idx in list_of_objects))


def xr_transpose_matrix_data(da, dim1, dim2) -> xr.DataArray:
    """Transpose data along two dimensions in an xarray DataArray.

    Parameters
    ----------
    da :
        xarray DataArray to transpose
    dim1 :
        name of the first dimension
    dim2 :
        name of the second dimension

    Returns
    -------
    xarray.DataArray
        xarray DataArray with transposed data at specified dimensions

    """
    i = swap_list_items(da.dims, dim1, dim2)

    return da.copy(data=da.transpose(*i).data)


def xr_matmul(
    a,
    b,
    dims_a,
    dims_b=None,
    dims_out=None,
    trans_a=False,
    trans_b=False,
    **apply_kwargs,
) -> xr.DataArray:
    """Calculate broadcasted np.matmul(a,b) for xarray objects.

    Should work for any size and shape of quadratic matrices contained in a DataArray.
    Ordering, broadcasting of dimensions should be taken care of by xarray internally.
    This can be used for both matrix * matrix and matrix * vector operations.

    Parameters
    ----------
    a :
        xarray object containing the first matrix
    b :
        xarray object containing the second matrix
    dims_a :
        name and order of dimensions in the first object
    dims_b :
        name and order of dimensions in the second object
        (if None, use dims_a) (Default value = None)
    dims_out :
        name and order of dimensions in the resulting object
        (if None, use dims_a unless dims_b has less items than dims_b)
        (Default value = None)
    trans_a :
        flag if matrix in a should be transposed (Default value = False)
    trans_b :
        flag if matrix in b should be transposed (Default value = False)
    **apply_kwargs :
        additional kwargs passed on to ur.apply_ufunc


    Returns
    -------
    xarray.DataArray

    """
    if dims_b is None:
        dims_b = dims_a
    if dims_out is None:
        if len(dims_a) <= len(dims_b):
            dims_out = dims_a
        else:
            dims_out = dims_b

    mul_func = np.matmul
    if len(dims_a) > len(dims_b):
        mul_func = mat_vec_mul

    if trans_a:
        dims_a = reversed(dims_a)
    if trans_b:
        dims_b = reversed(dims_b)

    return xr.apply_ufunc(
        mul_func,
        a,
        b,
        input_core_dims=[dims_a, dims_b],
        output_core_dims=[dims_out],
        **apply_kwargs,
    )


def xr_is_orthogonal_matrix(da: xr.DataArray, dims: List[str]) -> bool:
    """Check if  matrix along specific dimensions in a DataArray is orthogonal.

    TODO: make more general

    Parameters
    ----------
    da :
        xarray DataArray to test
    dims :
        list of dimensions along which to test

    Returns
    -------
    bool
        True if all matrices are orthogonal.

    """
    eye = np.eye(len(da.coords[dims[0]]), len(da.coords[dims[1]]))
    return np.allclose(xr_matmul(da, da, dims, trans_b=True), eye)


def xr_fill_all(da: xr.DataArray, order="bf") -> xr.DataArray:
    """Fill NaN values along all dimensions in xarray DataArray.

    Parameters
    ----------
    da : xarray.DataArray
        xarray object to fill
    order :
        order in which to apply bfill/ffill operation (Default value = "bf")

    Returns
    -------
    xarray.DataArray
        xarray object with NaN values filled in all dimensions

    """
    if order == "bf":
        for dim in da.dims:
            da = da.bfill(dim).ffill(dim)
    elif order == "fb":
        for dim in da.dims:
            da = da.ffill(dim).bfill(dim)
    else:
        raise ValueError(f"Order {order} is not supported (use 'bf' or 'fb)")
    return da


def xr_interp_like(
    da1: xr.DataArray,
    da2: Union[xr.DataArray, Dict[str, Any]],
    interp_coords: List[str] = None,
    broadcast_missing: bool = False,
    fillna: bool = True,
    method: str = "linear",
    assume_sorted: bool = False,
) -> xr.DataArray:
    """Interpolate DataArray along dimensions of another DataArray.

    Provides some utility options for handling out of range values and broadcasting.

    Parameters
    ----------
    da1 :
        xarray object with data to interpolate
    da2 :
        xarray or dict-like object along which dimensions to interpolate
    interp_coords :
        if not None, only interpolate along these coordinates of da2
        (Default value = None)
    broadcast_missing :
        broadcast da1 along all additional dimensions of da2
        (Default value = False)
    fillna :
        fill out of range NaN values
        (Default value = True)
    method :
        interpolation method to pass on to xarray.interp_like
        (Default value = "linear")
    assume_sorted :
        assume_sorted flag to pass on to xarray.interp_like
        (Default value = False)

    Returns
    -------
    xarray.DataArray
        interpolated DataArray

    """
    if isinstance(da2, (xr.DataArray, xr.Dataset)):
        sel_coords = da2.coords  # remember original interpolation coordinates
    else:  # assume da2 to be dict-like
        sel_coords = {
            k: (v if isinstance(v, Iterable) else [v]) for k, v in da2.items()
        }

    # store and strip pint units at this point, since the unit is lost during
    # interpolation and because of some other conflicts. Unit is restored before
    # returning the result.
    units = None
    if isinstance(da1.data, pint.Quantity):
        units = da1.data.units
        da1 = xr.DataArray(data=da1.data.magnitude, dims=da1.dims, coords=da1.coords)

    if interp_coords is not None:
        sel_coords = {k: v for k, v in sel_coords.items() if k in interp_coords}

    # create a new (empty) temporary dataset to use for interpolation
    # we need this if da2 is passed as an existing coordinate variable like origin.time
    da_temp = xr.DataArray(dims=sel_coords.keys(), coords=sel_coords)

    # make sure edge coordinate values of da1 are in new coordinate axis of da_temp
    if assume_sorted:
        # if all coordinates are sorted,we can use integer indexing for speedups
        edge_dict = {
            d: ([0, -1] if len(val) > 1 else [0])
            for d, val in da1.coords.items()
            if d in sel_coords
        }
        if len(edge_dict) > 0:
            da_temp = da_temp.combine_first(da1.isel(edge_dict))
    else:
        # select, combine with min/max values if coordinates not guaranteed to be sorted
        # maybe switch to idxmin()/idxmax() once it available
        # TODO: handle non-numeric dtypes ! currently cannot work on unsorted str types
        edge_dict = {
            d: ([val.min().data, val.max().data] if len(val) > 1 else [val.min().data])
            for d, val in da1.coords.items()
            if d in sel_coords
        }
        if len(edge_dict) > 0:
            da_temp = da_temp.combine_first(da1.sel(edge_dict))

    # handle singular dimensions in da1
    # TODO: should we handle coordinates or dimensions?
    singular_dims = [d for d in da1.coords if len(da1[d]) == 1]
    for dim in singular_dims:
        if dim in da_temp.coords:
            if len(da_temp.coords[dim]) > 1:
                if not fillna:
                    raise ValueError(
                        "Cannot use fillna=False with single point interpolation"
                    )
                exclude_dims = [d for d in da_temp.coords if not d == dim]
                # TODO: this always fills the dimension (inconsistent with fillna=False)
                da1 = xr_fill_all(da1.broadcast_like(da_temp, exclude=exclude_dims))
            else:
                del da_temp.coords[dim]

    # default interp_like will not add dimensions and fill out of range indexes with NaN
    da = da1.interp_like(da_temp, method=method, assume_sorted=assume_sorted)

    # copy original variable and coord attributes
    da.attrs = da1.attrs
    for key in da1.coords:
        da[key].attrs = da1[key].attrs

    # fill out of range nan values for all dimensions
    if fillna:
        da = xr_fill_all(da)

    if broadcast_missing:
        da = da.broadcast_like(da_temp)
    else:  # careful not to select coordinates that are only in da_temp
        sel_coords = {d: v for d, v in sel_coords.items() if d in da1.coords}

    result = da.sel(sel_coords)
    if units is not None:
        result = xr.DataArray(
            data=result.data * units, dims=result.dims, coords=result.coords,
        )

    return result


def _check_dtype(var_dtype, ref_dtype: dict) -> bool:
    """Check if dtype matches a reference dtype (or is subdtype).

    Parameters
    ----------
    var_dtype : numpy dtype
        A numpy-dtype to test against.
    ref_dtype : dict
        Python type or string description

    Returns
    -------
    bool
        True if dtypes matches.

    """
    if var_dtype != np.dtype(ref_dtype):
        if isinstance(ref_dtype, str):
            if (
                "timedelta64" in ref_dtype
                or "datetime64" in ref_dtype
                and np.issubdtype(var_dtype, np.dtype(ref_dtype))
            ):
                return True

        if not (
            np.issubdtype(var_dtype, np.dtype(ref_dtype)) and np.dtype(ref_dtype) == str
        ):
            return False

    return True


def xr_check_coords(dax: xr.DataArray, ref: dict) -> bool:
    """Validate the coordinates of the DataArray against a reference dictionary.

    The reference dictionary should have the dimensions as keys and those contain
    dictionaries with the following keywords (all optional):

    ``values``
        Specify exact coordinate values to match.

    ``dtype`` : str or type
        Ensure coordinate dtype matches at least one of the given dtypes.

    ``optional`` : boolean
        default ``False`` - if ``True``, the dimension has to be in the DataArray dax

    Parameters
    ----------
    dax : xarray.DataArray
        xarray object which should be validated
    ref : dict
        reference dictionary

    Returns
    -------
    bool
        True, if the test was a success, else an exception is raised

    Examples
    --------
    >>> import pandas as pd
    >>> import xarray as xr
    >>> import weldx as wx
    >>> dax = xr.DataArray(
    ...     data=np.ones((3, 2, 3)),
    ...     dims=["d1", "d2", "d3"],
    ...     coords={
    ...         "d1": np.array([-1, 0, 2], dtype=int),
    ...         "d2": pd.DatetimeIndex(["2020-05-01", "2020-05-03"]),
    ...         "d3": ["x", "y", "z"],
    ...     }
    ... )
    >>> ref = dict(
    ...     d1={"optional": True, "values": np.array([-1, 0, 2], dtype=int)},
    ...     d2={
    ...         "values": pd.DatetimeIndex(["2020-05-01", "2020-05-03"]),
    ...         "dtype": ["datetime64[ns]", "timedelta64[ns]"],
    ...     },
    ...     d3={"values": ["x", "y", "z"], "dtype": "<U1"},
    ... )
    >>> wx.utility.xr_check_coords(dax, ref)
    True

    """
    # only process the coords of the xarray
    if isinstance(dax, (xr.DataArray, xr.Dataset)):
        coords = dax.coords
    elif isinstance(
        dax,
        (
            xr.core.coordinates.DataArrayCoordinates,
            xr.core.coordinates.DatasetCoordinates,
        ),
    ):
        coords = dax
    else:
        raise ValueError("Input variable is not an xarray object")

    for key, check in ref.items():
        # check if the optional key is set to true
        if "optional" in check:
            if check["optional"] and key not in coords:
                # skip this key - it is not in dax
                continue

        if key not in coords:
            # Attributes not found in coords
            raise KeyError(f"Could not find required coordinate '{key}'.")

        # only if the key "values" is given do the validation
        if "values" in check:
            if not (coords[key].values == check["values"]).all():
                raise ValueError(f"Value mismatch in DataArray and ref['{key}']")

        # only if the key "dtype" is given do the validation
        if "dtype" in check:
            dtype_list = check["dtype"]
            if not isinstance(dtype_list, list):
                dtype_list = [dtype_list]
            if not any(
                _check_dtype(coords[key].dtype, var_dtype) for var_dtype in dtype_list
            ):
                raise TypeError(
                    f"Mismatch in the dtype of the DataArray and ref['{key}']"
                )

    return True


def xr_3d_vector(data, times=None) -> xr.DataArray:
    """Create an xarray 3d vector with correctly named dimensions and coordinates.

    Parameters
    ----------
    data :
        Data
    times :
        Optional time data (Default value = None)

    Returns
    -------
    xarray.DataArray

    """
    if times is not None:
        dsx = xr.DataArray(
            data=data, dims=["time", "c"], coords={"time": times, "c": ["x", "y", "z"]},
        )
    else:
        dsx = xr.DataArray(data=data, dims=["c"], coords={"c": ["x", "y", "z"]})
    return dsx.astype(float)


def xr_3d_matrix(data, times=None) -> xr.DataArray:
    """Create an xarray 3d matrix with correctly named dimensions and coordinates.

    Parameters
    ----------
    data :
        Data
    times :
        Optional time data (Default value = None)

    Returns
    -------
    xarray.DataArray

    """
    if times is not None:
        dsx = xr.DataArray(
            data=data,
            dims=["time", "c", "v"],
            coords={"time": times, "c": ["x", "y", "z"], "v": [0, 1, 2]},
        )
    else:
        dsx = xr.DataArray(
            data=data, dims=["c", "v"], coords={"c": ["x", "y", "z"], "v": [0, 1, 2]},
        )
    return dsx.astype(float)


def xr_interp_orientation_in_time(
    dsx: xr.DataArray, times: Union[pd.DatetimeIndex, pd.TimedeltaIndex]
) -> xr.DataArray:
    """Interpolate an xarray DataArray that represents orientation data in time.

    Parameters
    ----------
    dsx :
        xarray DataArray containing the orientation as matrix
    times :
        Time data

    Returns
    -------
    xarray.DataArray
        Interpolated data

    """
    if "time" not in dsx.coords:
        return dsx

    times = to_pandas_time_index(times)
    times_ds = to_pandas_time_index(dsx)
    time_ref = dsx.weldx.time_ref

    if len(times_ds) > 1:
        # extract intersecting times and add time range boundaries of the data set
        times_ds_limits = pd.Index([times_ds.min(), times_ds.max()])
        times_union = times.union(times_ds_limits)
        times_intersect = times_union[
            (times_union >= times_ds_limits[0]) & (times_union <= times_ds_limits[1])
        ]

        # interpolate rotations in the intersecting time range
        rotations_key = Rot.from_matrix(dsx.transpose(..., "c", "v").data)
        times_key = times_ds.astype(np.int64)
        rotations_interp = Slerp(times_key, rotations_key)(
            times_intersect.astype(np.int64)
        )
        dsx_out = xr_3d_matrix(rotations_interp.as_matrix(), times_intersect)
    else:
        # TODO: this case is not really well defined, maybe avoid?
        dsx_out = dsx

    # use interp_like to select original time values and correctly fill time dimension
    dsx_out = xr_interp_like(dsx_out, {"time": times}, fillna=True)

    # resync and reset to correct format
    if time_ref:
        dsx_out.weldx.time_ref = time_ref
    dsx_out = dsx_out.weldx.time_ref_restore()

    return dsx_out.transpose(..., "c", "v")


def xr_interp_coordinates_in_time(
    da: xr.DataArray, times: Union[pd.TimedeltaIndex, pd.DatetimeIndex]
) -> xr.DataArray:
    """Interpolate an xarray DataArray that represents 3d coordinates in time.

    Parameters
    ----------
    da : xarray.DataArray
        xarray DataArray
    times : pandas.TimedeltaIndex or pandas.DatetimeIndex
        Time data

    Returns
    -------
    xarray.DataArray
        Interpolated data

    """
    da = da.weldx.time_ref_unset()
    da = xr_interp_like(
        da, {"time": times}, assume_sorted=True, broadcast_missing=False, fillna=True
    )
    da = da.weldx.time_ref_restore()
    return da


def _as_valid_timestamp(value: Union[pd.Timestamp, str]) -> pd.Timestamp:
    """Create a valid (by convention) Timestamp object or raise TypeError.

    Parameters
    ----------
    value: pandas.Timestamp or str
        Value to convert to `pd.Timestamp`.

    Returns
    -------
    pandas.Timestamp

    """
    if isinstance(value, str):
        value = pd.Timestamp(value)
    if isinstance(value, pd.Timestamp):  # catch NaT from empty str.
        return value
    raise TypeError("Could not create a valid pandas.Timestamp.")


# weldx xarray Accessors --------------------------------------------------------


@xr.register_dataarray_accessor("weldx")
@xr.register_dataset_accessor("weldx")
class WeldxAccessor:
    """Custom accessor for extending DataArray functionality.

    See http://xarray.pydata.org/en/stable/internals.html#extending-xarray for details.
    """

    def __init__(self, xarray_obj):
        """Construct a WeldX xarray object."""
        self._obj = xarray_obj

    def interp_like(self, da, *args, **kwargs) -> xr.DataArray:  # pragma: no cover
        """Interpolate DataArray along dimensions of another DataArray.

        Provides some utility options for handling out of range values and broadcasting.
        See xr_interp_like for docstring and details.

        Returns
        -------
        xarray.DataArray
            interpolated DataArray

        """
        return xr_interp_like(self._obj, da, *args, **kwargs)

    def time_ref_unset(self) -> xr.DataArray:
        """Convert Timedelta + reference Timestamp to DatetimeIndex."""
        da = self._obj.copy()
        time_ref = da.weldx.time_ref
        if time_ref and is_timedelta64_dtype(da.time):
            da["time"] = da.time.data + time_ref
            da.time.attrs = self._obj.time.attrs  # restore old attributes !
        return da

    def time_ref_restore(self) -> xr.DataArray:
        """Convert DatetimeIndex back to TimedeltaIndex + reference Timestamp."""
        da = self._obj.copy()
        time_ref = da.weldx.time_ref
        if time_ref and is_datetime64_dtype(da.time):
            da["time"] = pd.DatetimeIndex(da.time.data) - time_ref
            da.time.attrs = self._obj.time.attrs  # restore old attributes !
        return da

    def reset_reference_time(self, time_ref_new: pd.Timestamp) -> xr.DataArray:
        """Return copy with time values shifted to new reference time."""
        da = self._obj.copy()
        da = da.weldx.time_ref_restore()
        da.weldx.time_ref = time_ref_new
        return da

    @property
    def time_ref(self) -> pd.Timestamp:
        """Get the time_ref value or `None` if not set."""
        da = self._obj
        if "time" in da.coords:
            if "time_ref" in da.time.attrs:
                return da.time.attrs["time_ref"]

        return None

    @time_ref.setter
    def time_ref(self, value: pd.Timestamp):
        """Convert INPLACE to new reference time.

        If no reference time exists, the new value will be assigned
        TODO: should None be allowed and pass through or raise TypeError ?
        """
        if "time" in self._obj.coords:
            value = _as_valid_timestamp(value)
            if self._obj.weldx.time_ref and is_timedelta64_dtype(self._obj.time):
                if value == self._obj.weldx.time_ref:
                    return
                _attrs = self._obj.time.attrs
                time_delta = value - self._obj.weldx.time_ref
                self._obj["time"] = self._obj.time.data - time_delta
                self._obj.time.attrs = _attrs  # restore old attributes !
                self._obj.time.attrs["time_ref"] = value  # set new time_ref value
            else:
                self._obj.time.attrs["time_ref"] = value



# Licensed under a 3-clause BSD style license - see LICENSE
# -*- coding: utf-8 -*-

import pint

from weldx.asdf.types import WeldxAsdfType
from weldx.constants import WELDX_QUANTITY as Q_

__all__ = ["PintQuantityType"]


class PintQuantityType(WeldxAsdfType):
    """A simple implementation of serializing a pint quantity as asdf quantity."""

    name = "unit/quantity"
    version = "1.1.0"
    types = [pint.Quantity]

    @classmethod
    def to_tree(cls, node: pint.Quantity, ctx):
        tree = {}
        value = node.magnitude
        if not value.shape:
            value = value.item()  # convert scalars to native Python numeric types
        tree["value"] = value
        tree["unit"] = str(node.units)
        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        value = tree["value"]
        unit = tree["unit"]
        quantity = Q_(value, unit)
        return quantity



"""Contains the serialization class for the weldx.core.TimeSeries."""

import numpy as np
import pint

from weldx.asdf.types import WeldxType
from weldx.constants import WELDX_QUANTITY as Q_
from weldx.core import MathematicalExpression, TimeSeries


class TimeSeriesTypeASDF(WeldxType):
    """Serialization class for weldx.core.TimeSeries"""

    name = "core/time_series"
    version = "1.0.0"
    types = [TimeSeries]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: TimeSeries, ctx):
        """
        Convert an 'weldx.core.TimeSeries' instance into YAML  representations.

        Parameters
        ----------
        node :
            Instance of the 'weldx.core.TimeSeries' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'weldx.core.TimeSeries'
            type to be serialized.

        """

        if isinstance(node.data, pint.Quantity):
            if node.shape == tuple([1]):
                return {
                    "unit": str(node.units),
                    "values": node.data.magnitude[0],
                }
            else:
                return {
                    "time": node.time,
                    "unit": str(node.units),
                    "shape": node.shape,
                    "interpolation": node.interpolation,
                    "values": node.data.magnitude,
                }
        return {"values": node.data, "unit": str(node.units), "shape": node.shape}

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into an 'weldx.core.TimeSeries'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        weldx.core.TimeSeries :
            An instance of the 'weldx.core.TimeSeries' type.

        """
        if not isinstance(tree["values"], MathematicalExpression):
            if "time" in tree:
                time = tree["time"]
                interpolation = tree["interpolation"]
            else:
                time = None
                interpolation = None
            values = Q_(np.asarray(tree["values"]), tree["unit"])
            return TimeSeries(values, time, interpolation)

        return TimeSeries(tree["values"])



"""Debug asdf save of groove implementation."""

import asdf

from weldx.asdf.extension import WeldxAsdfExtension, WeldxExtension
from weldx.asdf.tags.weldx.core.iso_groove import get_groove
from weldx.constants import WELDX_QUANTITY as Q_

v_groove = get_groove(
    groove_type="VGroove",
    workpiece_thickness=Q_(9, "mm"),
    groove_angle=Q_(50, "deg"),
    root_face=Q_(4, "mm"),
    root_gap=Q_(2, "mm"),
)
u_groove = get_groove(
    groove_type="UGroove",
    workpiece_thickness=Q_(15, "mm"),
    bevel_angle=Q_(9, "deg"),
    bevel_radius=Q_(6, "mm"),
    root_face=Q_(3, "mm"),
    root_gap=Q_(1, "mm"),
)
i_groove = get_groove(
    groove_type="IGroove", workpiece_thickness=Q_(4, "mm"), root_gap=Q_(4, "mm")
)
uv_groove = get_groove(
    groove_type="UVGroove",
    workpiece_thickness=Q_(12, "mm"),
    groove_angle=Q_(60, "deg"),
    bevel_angle=Q_(11, "deg"),
    bevel_radius=Q_(6, "mm"),
    root_face=Q_(4, "mm"),
    root_gap=Q_(2, "mm"),
)
vv_groove = get_groove(
    groove_type="VVGroove",
    workpiece_thickness=Q_(12, "mm"),
    groove_angle=Q_(70, "deg"),
    bevel_angle=Q_(13, "deg"),
    root_gap=Q_(3, "mm"),
    root_face=Q_(1, "mm"),
    root_face2=Q_(5, "mm"),
)
hv_groove = get_groove(
    groove_type="HVGroove",
    workpiece_thickness=Q_(9, "mm"),
    bevel_angle=Q_(55, "deg"),
    root_gap=Q_(2, "mm"),
    root_face=Q_(1, "mm"),
)
hu_groove = get_groove(
    groove_type="HUGroove",
    workpiece_thickness=Q_(18, "mm"),
    bevel_angle=Q_(15, "deg"),
    bevel_radius=Q_(8, "mm"),
    root_gap=Q_(2, "mm"),
    root_face=Q_(3, "mm"),
)
dv_groove = get_groove(
    groove_type="DoubleVGroove",
    workpiece_thickness=Q_(19, "mm"),
    groove_angle=Q_(40, "deg"),
    groove_angle2=Q_(60, "deg"),
    root_gap=Q_(2, "mm"),
    root_face=Q_(5, "mm"),
    root_face2=Q_(7, "mm"),
    root_face3=Q_(7, "mm"),
)
du_groove = get_groove(
    groove_type="DoubleUGroove",
    workpiece_thickness=Q_(33, "mm"),
    bevel_angle=Q_(8, "deg"),
    bevel_angle2=Q_(12, "deg"),
    bevel_radius=Q_(6, "mm"),
    bevel_radius2=Q_(6, "mm"),
    root_face=Q_(3, "mm"),
    root_face2=Q_(15, "mm"),
    root_gap=Q_(2, "mm"),
)
dhv_groove = get_groove(
    groove_type="DoubleHVGroove",
    workpiece_thickness=Q_(11, "mm"),
    bevel_angle=Q_(35, "deg"),
    bevel_angle2=Q_(60, "deg"),
    root_face2=Q_(5, "mm"),
    root_face=Q_(1, "mm"),
    root_gap=Q_(3, "mm"),
)
dhu_groove = get_groove(
    groove_type="DoubleHUGroove",
    workpiece_thickness=Q_(32, "mm"),
    bevel_angle=Q_(10, "deg"),
    bevel_angle2=Q_(20, "deg"),
    bevel_radius=Q_(8, "mm"),
    bevel_radius2=Q_(8, "mm"),
    root_face2=Q_(15, "mm"),
    root_face=Q_(2, "mm"),
    root_gap=Q_(2, "mm"),
)
ff_groove0 = get_groove(
    groove_type="FrontalFaceGroove", workpiece_thickness=Q_(5, "mm"), code_number="1.12"
)
ff_groove1 = get_groove(
    groove_type="FrontalFaceGroove",
    workpiece_thickness=Q_(5, "mm"),
    workpiece_thickness2=Q_(7, "mm"),
    groove_angle=Q_(80, "deg"),
    root_gap=Q_(1, "mm"),
    code_number="3.1.1",
)
ff_groove2 = get_groove(
    groove_type="FrontalFaceGroove",
    workpiece_thickness=Q_(2, "mm"),
    workpiece_thickness2=Q_(5, "mm"),
    root_gap=Q_(1, "mm"),
    code_number="3.1.2",
)
ff_groove3 = get_groove(
    groove_type="FrontalFaceGroove",
    workpiece_thickness=Q_(2, "mm"),
    workpiece_thickness2=Q_(5, "mm"),
    groove_angle=Q_(80, "deg"),
    root_gap=Q_(1, "mm"),
    code_number="3.1.3",
)
ff_groove4 = get_groove(
    groove_type="FrontalFaceGroove",
    workpiece_thickness=Q_(2, "mm"),
    workpiece_thickness2=Q_(5, "mm"),
    groove_angle=Q_(80, "deg"),
    special_depth=Q_(4, "mm"),
    code_number="4.1.2",
)
ff_groove5 = get_groove(
    groove_type="FrontalFaceGroove",
    workpiece_thickness=Q_(2, "mm"),
    workpiece_thickness2=Q_(5, "mm"),
    root_gap=Q_(1, "mm"),
    code_number="4.1.3",
)

target = asdf.AsdfFile(
    dict(
        test001=v_groove,
        test002=u_groove,
        test003=i_groove,
        test004=uv_groove,
        test005=vv_groove,
        test006=hv_groove,
        test007=hu_groove,
        test008=dv_groove,
        test009=du_groove,
        test010=dhv_groove,
        test011=dhu_groove,
        test012=ff_groove0,
        test013=ff_groove1,
        test014=ff_groove2,
        test015=ff_groove3,
        test016=ff_groove4,
        test017=ff_groove5,
    ),
    extensions=[WeldxAsdfExtension(), WeldxExtension()],
)
target.write_to("testfile.yml", all_array_storage="inline")



from xarray import Dataset

import weldx.asdf.tags.weldx.core.common_types as ct
from weldx.asdf.types import WeldxType


class XarrayDatasetASDF(WeldxType):
    """Serialization class for xarray.Dataset"""

    name = "core/dataset"
    version = "1.0.0"
    types = [Dataset]
    requires = ["weldx"]
    handle_dynamic_subclasses = True

    @classmethod
    def to_tree(cls, node: Dataset, ctx):
        """
        Convert an instance of the 'xarray.Dataset' type into YAML representations.

        Parameters
        ----------
        node :
            Instance of the 'xarray.Dataset' type to be serialized.

        ctx :
            An instance of the 'AsdfFile' object that is being written out.

        Returns
        -------
            A basic YAML type ('dict', 'list', 'str', 'int', 'float', or
            'complex') representing the properties of the 'xarray.Dataset' type to be
            serialized.

        """
        attributes = node.attrs
        coordinates = []
        dimensions = []
        variables = []

        for name, length in dict(node.dims).items():
            dimensions.append(ct.Dimension(name, length))

        for name, data in node.data_vars.items():
            variables.append(ct.Variable(name, data.dims, data.data))

        for name, data in node.coords.items():
            coordinates.append(ct.Variable(name, data.dims, data.data))

        tree = {
            "attributes": attributes,
            "coordinates": coordinates,
            "dimensions": dimensions,
            "variables": variables,
        }

        # variables = []
        # for variable_name in node:
        #    variables.append(netcdf.NetCDFVariable(node[variable_name]))
        # tree = {"coordinates": dict(node.coords), "variables": variables}

        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """
        Converts basic types representing YAML trees into an 'xarray.Dataset'.

        Parameters
        ----------
        tree :
            An instance of a basic Python type (possibly nested) that
            corresponds to a YAML subtree.
        ctx :
            An instance of the 'AsdfFile' object that is being constructed.

        Returns
        -------
        xarray.Dataset :
            An instance of the 'xarray.Dataset' type.

        """
        data_vars = {}

        for variable in tree["variables"]:
            data_vars[variable.name] = (variable.dimensions, variable.data)

        coords = {}
        for coordinate in tree["coordinates"]:
            coords[coordinate.name] = (coordinate.dimensions, coordinate.data)

        obj = Dataset(data_vars=data_vars, coords=coords)

        obj.attrs = tree["attributes"]

        return obj



"""Define constants for global library use."""

from pint import UnitRegistry

WELDX_UNIT_REGISTRY = UnitRegistry(
    preprocessors=[lambda string: string.replace("%", "percent")],  # allow %-sign
    force_ndarray_like=True,
)
WELDX_QUANTITY = WELDX_UNIT_REGISTRY.Quantity

# add percent unit
WELDX_UNIT_REGISTRY.define("percent = 0.01*count = %")
# swap plank constant for hour definition
WELDX_UNIT_REGISTRY.define("hour = 60*minute = h = hr")



import matplotlib.pyplot as plt

from weldx.transformations import CoordinateSystemManager

csm_global = CoordinateSystemManager("root", "global coordinate systems")
csm_global.create_cs("specimen", "root", coordinates=[1, 2, 3])
csm_global.create_cs("robot head", "root", coordinates=[4, 5, 6])

csm_specimen = CoordinateSystemManager("specimen", "specimen coordinate systems")
csm_specimen.create_cs("thermo couple 1", "specimen", coordinates=[1, 1, 0])
csm_specimen.create_cs("thermo couple 2", "specimen", coordinates=[1, 4, 0])

csm_robot = CoordinateSystemManager("robot head", "robot coordinate systems")
csm_robot.create_cs("torch", "robot head", coordinates=[0, 0, -2])
csm_robot.create_cs("mount point 1", "robot head", coordinates=[0, 1, -1])
csm_robot.create_cs("mount point 2", "robot head", coordinates=[0, -1, -1])

csm_scanner = CoordinateSystemManager("scanner", "scanner coordinate systems")
csm_scanner.create_cs("mount point 1", "scanner", coordinates=[0, 0, 2])

csm_robot.merge(csm_scanner)
csm_global.merge(csm_robot)
csm_global.merge(csm_specimen)
csm_global.plot()
plt.show()



"""Contains measurement related classes and functions."""

from dataclasses import dataclass, field
from typing import List, Union  # noqa: F401

import xarray as xr


# measurement --------------------------------------------------------------------------
@dataclass
class Data:
    """Simple dataclass implementation for measurement data."""

    name: str
    data: xr.DataArray


@dataclass
class Error:
    """Simple dataclass implementation for signal transformation errors."""

    deviation: float


@dataclass
class Signal:
    """Simple dataclass implementation for measurement signals."""

    signal_type: str
    unit: str
    data: Union[Data, None]


@dataclass
class DataTransformation:
    """Simple dataclass implementation for signal transformations."""

    name: str
    input_signal: Signal
    output_signal: Signal
    error: Error
    func: str = None
    meta: str = None


@dataclass
class Source:
    """Simple dataclass implementation for signal sources."""

    name: str
    output_signal: Signal
    error: Error


@dataclass
class MeasurementChain:
    """Simple dataclass implementation for measurement chains."""

    name: str
    data_source: Source
    data_processors: List = field(default_factory=lambda: [])


@dataclass
class Measurement:
    """Simple dataclass implementation for generic measurements."""

    name: str
    data: Data
    measurement_chain: MeasurementChain


# equipment ----------------------------------------------------------------------------
@dataclass
class GenericEquipment:
    """Simple dataclass implementation for generic equipment."""

    name: str
    sources: List = field(default_factory=lambda: [])
    data_transformations: List = field(default_factory=lambda: [])



# Licensed under a 3-clause BSD style license - see LICENSE
# -*- coding: utf-8 -*-

from typing import List

import numpy as np
import pandas as pd
from asdf.tagged import TaggedDict

from weldx.asdf.types import WeldxType

__all__ = ["TimedeltaIndexType"]


class TimedeltaIndexType(WeldxType):
    """A simple implementation of serializing pandas TimedeltaIndex."""

    name = "time/timedeltaindex"
    version = "1.0.0"
    types = [pd.TimedeltaIndex]

    @classmethod
    def to_tree(cls, node: pd.TimedeltaIndex, ctx):
        """Serialize TimedeltaIndex to tree."""
        tree = {}
        if node.inferred_freq is not None:
            tree["freq"] = node.inferred_freq
        else:
            tree["values"] = node.values.astype(np.int64)

        tree["start"] = node[0]
        tree["end"] = node[-1]
        tree["min"] = node.min()
        tree["max"] = node.max()

        return tree

    @classmethod
    def from_tree(cls, tree, ctx):
        """Construct TimedeltaIndex from tree."""
        if "freq" in tree:
            return pd.timedelta_range(
                start=tree["start"], end=tree["end"], freq=tree["freq"]
            )
        values = tree["values"]
        return pd.TimedeltaIndex(values)

    @staticmethod
    def shape_from_tagged(tree: TaggedDict) -> List[int]:
        """Calculate the shape (length of TDI) from static tagged tree instance."""
        if "freq" in tree:
            tdi_temp = pd.timedelta_range(
                start=tree["start"]["value"],
                end=tree["end"]["value"],
                freq=tree["freq"],
            )
            return [len(tdi_temp)]
        else:
            return tree["values"]["shape"]



# Licensed under a 3-clause BSD style license - see LICENSE
# -*- coding: utf-8 -*-

from asdf.extension import AsdfExtension, BuiltinExtension

from weldx.asdf.constants import WELDX_SCHEMA_URI_BASE, WELDX_URL_MAPPING

from .types import _weldx_asdf_types, _weldx_types

__all__ = ["WeldxExtension", "WeldxAsdfExtension"]


class WxSyntaxError(Exception):
    """Exception raising on custom weldx ASDF syntax errors."""


class WeldxExtension(AsdfExtension):
    """Extension class registering types with both tags and schemas defined by weldx."""

    @property
    def types(self):
        # There are no types yet!
        return _weldx_types

    @property
    def tag_mapping(self):
        return [("tag:weldx.bam.de:weldx", WELDX_SCHEMA_URI_BASE + "weldx{tag_suffix}")]

    @property
    def url_mapping(self):
        return WELDX_URL_MAPPING


class WeldxAsdfExtension(BuiltinExtension):
    """This extension is used to register custom tag types that have schemas defined
    by ASDF, but have tag implementations defined in the weldx package

    """

    @property
    def types(self):
        return _weldx_asdf_types
